package java.util.concurrent;

import java.util.concurrent.locks.LockSupport;
import sun.misc.Unsafe;
import java.util.concurrent.atomic.AtomicReference;

public class Phaser
{
    private volatile long state;
    private static final int MAX_PARTIES = 65535;
    private static final int MAX_PHASE = Integer.MAX_VALUE;
    private static final int PARTIES_SHIFT = 16;
    private static final int PHASE_SHIFT = 32;
    private static final int UNARRIVED_MASK = 65535;
    private static final long PARTIES_MASK = 4294901760L;
    private static final long COUNTS_MASK = 4294967295L;
    private static final long TERMINATION_BIT = Long.MIN_VALUE;
    private static final int ONE_ARRIVAL = 1;
    private static final int ONE_PARTY = 65536;
    private static final int ONE_DEREGISTER = 65537;
    private static final int EMPTY = 1;
    private final Phaser parent;
    private final Phaser root;
    private final AtomicReference<Phaser.QNode> evenQ;
    private final AtomicReference<Phaser.QNode> oddQ;
    private static final int NCPU;
    static final int SPINS_PER_ARRIVAL;
    private static final Unsafe UNSAFE;
    private static final long stateOffset;
    
    private static int unarrivedOf(long n) {
        int n2 = (int)n;
        return (n2 == 1) ? 0 : (n2 & 0xFFFF);
    }
    
    private static int partiesOf(long n) {
        return (int)n >>> 16;
    }
    
    private static int phaseOf(long n) {
        return (int)(n >>> 32);
    }
    
    private static int arrivedOf(long n) {
        int n2 = (int)n;
        return (n2 == 1) ? 0 : ((n2 >>> 16) - (n2 & 0xFFFF));
    }
    
    private AtomicReference<Phaser.QNode> queueFor(int n) {
        return ((n & 0x1) == 0x0) ? this.evenQ : this.oddQ;
    }
    
    private String badArrive(long n) {
        return "Attempted arrival of unregistered party for " + this.stateToString(n);
    }
    
    private String badRegister(long n) {
        return "Attempt to register more than 65535 parties for " + this.stateToString(n);
    }
    
    private int doArrive(int n) {
        Phaser root = this.root;
        while (true) {
            long expected = (root == this) ? this.state : this.reconcileState();
            int n2 = (int)(expected >>> 32);
            if (n2 < 0) {
                return n2;
            }
            int n3 = (int)expected;
            int n4 = (n3 == 1) ? 0 : (n3 & 0xFFFF);
            if (n4 <= 0) {
                throw new IllegalStateException(this.badArrive(expected));
            }
            long n5;
            if (Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, expected, n5 = expected - (long)n)) {
                if (n4 == 1) {
                    long n6 = n5 & 0xFFFF0000L;
                    int n7 = (int)n6 >>> 16;
                    if (root == this) {
                        long n8;
                        if (this.onAdvance(n2, n7)) {
                            n8 = (n6 | Long.MIN_VALUE);
                        }
                        else if (n7 == 0) {
                            n8 = (n6 | 0x1L);
                        }
                        else {
                            n8 = (n6 | (long)n7);
                        }
                        Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, n5, n8 | (long)(n2 + 1 & Integer.MAX_VALUE) << 32);
                        this.releaseWaiters(n2);
                    }
                    else if (n7 == 0) {
                        n2 = this.parent.doArrive(65537);
                        Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, n5, n5 | 0x1L);
                    }
                    else {
                        n2 = this.parent.doArrive(1);
                    }
                }
                return n2;
            }
            else {
                continue;
            }
        }
    }
    
    private int doRegister(int n) {
        long n2 = (long)n << 16 | (long)n;
        Phaser parent = this.parent;
        int doRegister;
        while (true) {
            long state = (parent == null) ? this.state : this.reconcileState();
            int n3 = (int)state;
            int n4 = n3 >>> 16;
            int n5 = n3 & 0xFFFF;
            if (n > 65535 - n4) {
                throw new IllegalStateException(this.badRegister(state));
            }
            doRegister = (int)(state >>> 32);
            if (doRegister < 0) {
                break;
            }
            if (n3 != 1) {
                if (parent == null || this.reconcileState() == state) {
                    if (n5 == 0) {
                        this.root.internalAwaitAdvance(doRegister, null);
                    }
                    else {
                        if (Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, state, state + n2)) {
                            break;
                        }
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
            else if (parent == null) {
                if (Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, state, (long)doRegister << 32 | n2)) {
                    break;
                }
                continue;
            }
            else {
                synchronized (this) {
                    if (this.state == state) {
                        doRegister = parent.doRegister(1);
                        if (doRegister < 0) {
                            break;
                        }
                        while (!Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, state, (long)doRegister << 32 | n2)) {
                            state = this.state;
                            doRegister = (int)(this.root.state >>> 32);
                        }
                        break;
                    }
                    else {
                        continue;
                    }
                }
            }
        }
        return doRegister;
    }
    
    private long reconcileState() {
        Phaser root = this.root;
        long expected = this.state;
        if (root != this) {
            int n;
            int n2;
            while ((n = (int)(root.state >>> 32)) != (int)(expected >>> 32) && !Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, expected, expected = ((long)n << 32 | ((n < 0) ? (expected & 0xFFFFFFFFL) : (((n2 = (int)expected >>> 16) == 0) ? 1L : ((expected & 0xFFFF0000L) | (long)n2)))))) {
                expected = this.state;
            }
        }
        return expected;
    }
    
    public Phaser() {
        this(null, 0);
    }
    
    public Phaser(int n) {
        this(null, n);
    }
    
    public Phaser(Phaser phaser) {
        this(phaser, 0);
    }
    
    public Phaser(Phaser parent, int n) {
        if (n >>> 16 != 0) {
            throw new IllegalArgumentException("Illegal number of parties");
        }
        int doRegister = 0;
        if ((this.parent = parent) != null) {
            Phaser root = parent.root;
            this.root = root;
            this.evenQ = root.evenQ;
            this.oddQ = root.oddQ;
            if (n != 0) {
                doRegister = parent.doRegister(1);
            }
        }
        else {
            this.root = this;
            this.evenQ = new AtomicReference();
            this.oddQ = new AtomicReference();
        }
        this.state = ((n == 0) ? 1L : ((long)doRegister << 32 | (long)n << 16 | (long)n));
        return;
    }
    
    public int register() {
        return this.doRegister(1);
    }
    
    public int bulkRegister(int n) {
        if (n < 0) {
            throw new IllegalArgumentException();
        }
        if (n == 0) {
            return this.getPhase();
        }
        return this.doRegister(n);
    }
    
    public int arrive() {
        return this.doArrive(1);
    }
    
    public int arriveAndDeregister() {
        return this.doArrive(65537);
    }
    
    public int arriveAndAwaitAdvance() {
        Phaser root = this.root;
        while (true) {
            long expected = (root == this) ? this.state : this.reconcileState();
            int n = (int)(expected >>> 32);
            if (n < 0) {
                return n;
            }
            int n2 = (int)expected;
            int n3 = (n2 == 1) ? 0 : (n2 & 0xFFFF);
            if (n3 <= 0) {
                throw new IllegalStateException(this.badArrive(expected));
            }
            long expected2;
            if (Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, expected, expected2 = expected - 1L)) {
                if (n3 > 1) {
                    return root.internalAwaitAdvance(n, null);
                }
                if (root != this) {
                    return this.parent.arriveAndAwaitAdvance();
                }
                long n4 = expected2 & 0xFFFF0000L;
                int n5 = (int)n4 >>> 16;
                long n6;
                if (this.onAdvance(n, n5)) {
                    n6 = (n4 | Long.MIN_VALUE);
                }
                else if (n5 == 0) {
                    n6 = (n4 | 0x1L);
                }
                else {
                    n6 = (n4 | (long)n5);
                }
                int n7 = n + 1 & Integer.MAX_VALUE;
                if (!Phaser.UNSAFE.compareAndSwapLong(this, Phaser.stateOffset, expected2, n6 | (long)n7 << 32)) {
                    return (int)(this.state >>> 32);
                }
                this.releaseWaiters(n);
                return n7;
            }
            else {
                continue;
            }
        }
    }
    
    public int awaitAdvance(int n) {
        Phaser root = this.root;
        int n2 = (int)(((root == this) ? this.state : this.reconcileState()) >>> 32);
        if (n < 0) {
            return n;
        }
        if (n2 == n) {
            return root.internalAwaitAdvance(n, null);
        }
        return n2;
    }
    
    public int awaitAdvanceInterruptibly(int n) throws InterruptedException {
        Phaser root = this.root;
        int internalAwaitAdvance = (int)(((root == this) ? this.state : this.reconcileState()) >>> 32);
        if (n < 0) {
            return n;
        }
        if (internalAwaitAdvance == n) {
            Phaser.QNode qNode = new Phaser.QNode(this, n, true, false, 0L);
            internalAwaitAdvance = root.internalAwaitAdvance(n, qNode);
            if (qNode.wasInterrupted) {
                throw new InterruptedException();
            }
        }
        return internalAwaitAdvance;
    }
    
    public int awaitAdvanceInterruptibly(int n, long n2, TimeUnit timeUnit) throws InterruptedException, TimeoutException {
        long nanos = timeUnit.toNanos(n2);
        Phaser root = this.root;
        int internalAwaitAdvance = (int)(((root == this) ? this.state : this.reconcileState()) >>> 32);
        if (n < 0) {
            return n;
        }
        if (internalAwaitAdvance == n) {
            Phaser.QNode qNode = new Phaser.QNode(this, n, true, true, nanos);
            internalAwaitAdvance = root.internalAwaitAdvance(n, qNode);
            if (qNode.wasInterrupted) {
                throw new InterruptedException();
            }
            if (internalAwaitAdvance == n) {
                throw new TimeoutException();
            }
        }
        return internalAwaitAdvance;
    }
    
    public void forceTermination() {
        Phaser root = this.root;
        long state;
        while ((state = root.state) >= 0L) {
            if (!Phaser.UNSAFE.compareAndSwapLong(root, Phaser.stateOffset, state, state | Long.MIN_VALUE)) {
                continue;
            }
            this.releaseWaiters(0);
            this.releaseWaiters(1);
            return;
        }
    }
    
    public final int getPhase() {
        return (int)(this.root.state >>> 32);
    }
    
    public int getRegisteredParties() {
        return partiesOf(this.state);
    }
    
    public int getArrivedParties() {
        return arrivedOf(this.reconcileState());
    }
    
    public int getUnarrivedParties() {
        return unarrivedOf(this.reconcileState());
    }
    
    public Phaser getParent() {
        return this.parent;
    }
    
    public Phaser getRoot() {
        return this.root;
    }
    
    public boolean isTerminated() {
        return this.root.state < 0L;
    }
    
    protected boolean onAdvance(int n, int n2) {
        return n2 == 0;
    }
    
    @Override
    public String toString() {
        return this.stateToString(this.reconcileState());
    }
    
    private String stateToString(long n) {
        return super.toString() + "[phase = " + phaseOf(n) + " parties = " + partiesOf(n) + " arrived = " + arrivedOf(n) + "]";
    }
    
    private void releaseWaiters(int n) {
        AtomicReference atomicReference = ((n & 0x1) == 0x0) ? this.evenQ : this.oddQ;
        Phaser.QNode expectedValue;
        while ((expectedValue = (Phaser.QNode)atomicReference.get()) != null && expectedValue.phase != (int)(this.root.state >>> 32)) {
            Thread thread;
            if (!atomicReference.compareAndSet(expectedValue, expectedValue.next) || (thread = expectedValue.thread) == null) {
                continue;
            }
            expectedValue.thread = null;
            LockSupport.unpark(thread);
        }
    }
    
    private int abortWait(int n) {
        AtomicReference atomicReference = ((n & 0x1) == 0x0) ? this.evenQ : this.oddQ;
        int n2;
        while (true) {
            Phaser.QNode expectedValue = (Phaser.QNode)atomicReference.get();
            n2 = (int)(this.root.state >>> 32);
            Thread thread;
            if (expectedValue == null || ((thread = expectedValue.thread) != null && expectedValue.phase == n2)) {
                break;
            }
            if (atomicReference.compareAndSet(expectedValue, expectedValue.next) && thread != null) {
                expectedValue.thread = null;
                LockSupport.unpark(thread);
            }
            else {
                continue;
            }
        }
        return n2;
    }
    
    private int internalAwaitAdvance(int n, Phaser.QNode newValue) {
        this.releaseWaiters(n - 1);
        boolean compareAndSet = false;
        int n2 = 0;
        int spins_PER_ARRIVAL = Phaser.SPINS_PER_ARRIVAL;
        long state;
        int n3;
        while ((n3 = (int)((state = this.state) >>> 32)) == n) {
            if (newValue == null) {
                int n4 = (int)state & 0xFFFF;
                if (n4 != n2 && (n2 = n4) < Phaser.NCPU) {
                    spins_PER_ARRIVAL += Phaser.SPINS_PER_ARRIVAL;
                }
                int interrupted = Thread.interrupted();
                if (interrupted || --spins_PER_ARRIVAL < 0) {
                    newValue = new Phaser.QNode(this, n, false, false, 0L);
                    newValue.wasInterrupted = interrupted;
                }
                else {
                    continue;
                }
            }
            else {
                if (newValue.isReleasable()) {
                    break;
                }
                if (compareAndSet == 0) {
                    AtomicReference atomicReference = ((n & 0x1) == 0x0) ? this.evenQ : this.oddQ;
                    Phaser.QNode next = (Phaser.QNode)atomicReference.get();
                    newValue.next = next;
                    Phaser.QNode expectedValue = next;
                    if ((expectedValue == null || expectedValue.phase == n) && (int)(this.state >>> 32) == n) {
                        compareAndSet = atomicReference.compareAndSet(expectedValue, newValue);
                    }
                    else {
                        continue;
                    }
                }
                else {
                    try {
                        ForkJoinPool.managedBlock(newValue);
                    }
                    catch (InterruptedException ex) {
                        newValue.wasInterrupted = true;
                    }
                }
            }
        }
        if (newValue != null) {
            if (newValue.thread != null) {
                newValue.thread = null;
            }
            if (newValue.wasInterrupted && !newValue.interruptible) {
                Thread.currentThread().interrupt();
            }
            if (n3 == n && (n3 = (int)(this.state >>> 32)) == n) {
                return this.abortWait(n);
            }
        }
        this.releaseWaiters(n);
        return n3;
    }
    
    static {
        NCPU = Runtime.getRuntime().availableProcessors();
        SPINS_PER_ARRIVAL = ((Phaser.NCPU < 2) ? 1 : 256);
        try {
            UNSAFE = Unsafe.getUnsafe();
            stateOffset = Phaser.UNSAFE.objectFieldOffset(Phaser.class.getDeclaredField("state"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class QNode implements ForkJoinPool.ManagedBlocker
    {
        final Phaser phaser;
        final int phase;
        final boolean interruptible;
        final boolean timed;
        boolean wasInterrupted;
        long nanos;
        final long deadline;
        volatile Thread thread;
        Phaser.QNode next;
        
        QNode(Phaser phaser, int phase, boolean interruptible, boolean timed, long nanos) {
            this.phaser = phaser;
            this.phase = phase;
            this.interruptible = interruptible;
            this.nanos = nanos;
            this.timed = timed;
            this.deadline = (timed ? (System.nanoTime() + nanos) : 0L);
            this.thread = Thread.currentThread();
        }
        
        @Override
        public boolean isReleasable() {
            if (this.thread == null) {
                return true;
            }
            if (this.phaser.getPhase() != this.phase) {
                this.thread = null;
                return true;
            }
            if (Thread.interrupted()) {
                this.wasInterrupted = true;
            }
            if (this.wasInterrupted && this.interruptible) {
                this.thread = null;
                return true;
            }
            if (this.timed) {
                if (this.nanos > 0L) {
                    this.nanos = this.deadline - System.nanoTime();
                }
                if (this.nanos <= 0L) {
                    this.thread = null;
                    return true;
                }
            }
            return false;
        }
        
        @Override
        public boolean block() {
            if (this.isReleasable()) {
                return true;
            }
            if (!this.timed) {
                LockSupport.park(this);
            }
            else if (this.nanos > 0L) {
                LockSupport.parkNanos(this, this.nanos);
            }
            return this.isReleasable();
        }
    }
}
